{-# LANGUAGE OverloadedStrings #-}

module Test.Shift.ParsersSpec where

import Data.HashSet (fromList)
import Shift (BreakingChange (BreakingChange))
import Shift.Parsers
  ( breakingChange,
    commit,
    commitType,
    manyCharsTill,
    someCharsTill,
    spaced,
  )
import Shift.Types
  ( CommitType (CTFeature),
    ConventionalCommit
      ( ConventionalCommit,
        _ccAffectedTickets,
        _ccBody,
        _ccBreakingChanges,
        _ccScope,
        _ccSubject,
        _ccType
      ),
    ParsedCommit (PCConventional),
  )
import Test.Hspec (Spec, describe, it)
import Test.Hspec.Megaparsec (shouldFailOn, shouldParse)
import Text.Megaparsec (parse)
import Text.Megaparsec.Char (eol, string)

spec_Parsers :: Spec
spec_Parsers = do
  -- spaced
  describe "spaced" $ do
    it "parses using the inner parser" $ do
      parse (spaced $ string "hello") "" "hello" `shouldParse` "hello"
    it "parses using the inner parser (with whitespace)" $ do
      parse (spaced $ string "hello") "" "   hello" `shouldParse` "hello"
    it "parses using the inner parser (with newlines)" $ do
      parse (spaced $ string "hello") "" "\n\n\nhello" `shouldParse` "hello"
    it "parses using the inner parser (with newlines and whitespace)" $ do
      parse (spaced $ string "hello") "" "\n  \n \nhello" `shouldParse` "hello"
  -- manyCharsTill
  describe "manyCharsTill" $ do
    it "parses empty strings (with EOL)" $ do
      parse (manyCharsTill eol) "" "\n" `shouldParse` ""
    it "parses strings (with EOL)" $ do
      parse (manyCharsTill eol) "" "hello\n" `shouldParse` "hello"
    it "fails on empty strings (without EOL)" $ do
      parse (manyCharsTill eol) "" `shouldFailOn` ""
  -- someCharsTill
  describe "someCharsTill" $ do
    it "fails on empty strings (with EOL)" $ do
      parse (someCharsTill eol) "" `shouldFailOn` "\n"
    it "parses strings (with EOL)" $ do
      parse (someCharsTill eol) "" "hello\n" `shouldParse` "hello"
    it "fails on empty strings (without EOL)" $ do
      parse (someCharsTill eol) "" `shouldFailOn` ""
  -- commitType
  describe "commitType" $ do
    it "parses a commit type" $ do
      parse commitType "" "feat" `shouldParse` CTFeature
    it "should fail if the commit type is unknown" $ do
      parse commitType "" `shouldFailOn` "omg"
  -- breakingChange
  describe "breakingChange" $ do
    it "parses a breaking change without a body" $ do
      parse
        breakingChange
        ""
        "BREAKING CHANGE: This will require a new OS upgrade.\n"
        `shouldParse` BreakingChange "This will require a new OS upgrade." Nothing
    it "parses a breaking change with a body" $ do
      parse
        breakingChange
        ""
        "BREAKING CHANGE: This will require a new OS upgrade.\n\
        \\n\
        \Go to example.org to download a new OS."
        `shouldParse` BreakingChange
          "This will require a new OS upgrade."
          (Just "Go to example.org to download a new OS.")
  -- commit
  describe "commit" $ do
    it "parses a conventional commit" $ do
      parse commit "" "feat(src): Add README file"
        `shouldParse` ( PCConventional $
                          ConventionalCommit
                            { _ccType = CTFeature,
                              _ccScope = Just "src",
                              _ccSubject = "Add README file",
                              _ccBody = "",
                              _ccBreakingChanges = [],
                              _ccAffectedTickets = []
                            }
                      )
    it "parses a conventional commit (without scope)" $ do
      parse commit "" "feat: Add README file"
        `shouldParse` ( PCConventional $
                          ConventionalCommit
                            { _ccType = CTFeature,
                              _ccScope = Nothing,
                              _ccSubject = "Add README file",
                              _ccBody = "",
                              _ccBreakingChanges = [],
                              _ccAffectedTickets = []
                            }
                      )
    it "parses a conventional commit with a body" $ do
      parse
        commit
        ""
        "feat(Git): Render empty changelogs\n\
        \\n\
        \Adds support for empty diffs. Instead of throwing/erroring out, `shift` will\n\
        \now render an empty changelog when the diff between two refs is empty.\n"
        `shouldParse` ( PCConventional $
                          ConventionalCommit
                            { _ccType = CTFeature,
                              _ccScope = Just "Git",
                              _ccSubject = "Render empty changelogs",
                              _ccBody =
                                "Adds support for empty diffs. Instead of throwing/erroring out, `shift` will\n\
                                \now render an empty changelog when the diff between two refs is empty.",
                              _ccBreakingChanges = [],
                              _ccAffectedTickets = []
                            }
                      )
    it "parses a conventional commit with breaking changes and a body" $ do
      parse
        commit
        ""
        "feat(src): Add README file\n\
        \\n\
        \This is an example commit body.\n\
        \\n\
        \BREAKING CHANGE: This will require a new OS upgrade.\n\
        \\n\
        \Go to example.org to download a new OS.\n"
        `shouldParse` ( PCConventional $
                          ConventionalCommit
                            { _ccType = CTFeature,
                              _ccScope = Just "src",
                              _ccSubject = "Add README file",
                              _ccBody = "This is an example commit body.",
                              _ccBreakingChanges =
                                [ BreakingChange
                                    "This will require a new OS upgrade."
                                    (Just "Go to example.org to download a new OS.")
                                ],
                              _ccAffectedTickets = []
                            }
                      )
    it "parses a conventional commit (kitchen sink)" $ do
      parse
        commit
        ""
        "feat(src): Add README file\n\
        \\n\
        \This is an example commit body.\n\
        \\n\
        \BREAKING CHANGE: This will require a new OS upgrade.\n\
        \\n\
        \Go to example.org to download a new OS.\n\
        \\n\
        \Fixes: #123, #234\n"
        `shouldParse` ( PCConventional $
                          ConventionalCommit
                            { _ccType = CTFeature,
                              _ccScope = Just "src",
                              _ccSubject = "Add README file",
                              _ccBody = "This is an example commit body.",
                              _ccBreakingChanges =
                                [ BreakingChange
                                    "This will require a new OS upgrade."
                                    (Just "Go to example.org to download a new OS.")
                                ],
                              _ccAffectedTickets =
                                [ fromList
                                    [ ("Fixes", "234"),
                                      ("Fixes", "123")
                                    ]
                                ]
                            }
                      )
